# -*- coding: utf-8; -*-
"""add label_profile.sync_me

Revision ID: f7dc4e92752c
Revises: 648cb088658c
Create Date: 2018-12-04 18:19:19.927787

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'f7dc4e92752c'
down_revision = '648cb088658c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # label_profile
    op.add_column('label_profile', sa.Column('sync_me', sa.Boolean(), nullable=True))
    op.add_column('label_profile_version', sa.Column('sync_me', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # label_profile
    op.drop_column('label_profile_version', 'sync_me')
    op.drop_column('label_profile', 'sync_me')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for Equal Exchange
"""

from __future__ import unicode_literals, absolute_import

import re
import decimal

import six
import xlrd

from rattail.db import model
from rattail.gpc import GPC
from rattail.vendors.catalogs import CatalogParser


class EqualExchangeCatalogParser(CatalogParser):
    """
    Vendor catalog parser for Equal Exchange.
    """
    key = 'rattail.contrib.equalexchange'
    display = "Equal Exchange"
    vendor_key = 'equal-exchange'

    def parse_rows(self, path, progress=None):
        book = xlrd.open_workbook(path)

        # Bulk
        sheet = book.sheet_by_index(0)
        for row in self.parse_sheet_rows(sheet, "Bulk", progress=progress):
            yield row

        # Grocery
        sheet = book.sheet_by_index(1)
        for row in self.parse_sheet_rows(sheet, "Grocery", progress=progress):
            yield row

    def parse_sheet_rows(self, sheet, department_name, progress=None):
        unit_price_pattern = re.compile(r'^\$(\d+\.\d\d)/lb$')

        for r in range(sheet.nrows):
            row = model.VendorCatalogBatchRow()
            row.brand_name = "Equal Exchange"
            row.department_name = department_name

            # column A (ITEM DESCRIPTION)
            row.description = sheet.cell_value(r, 0)

            # column B (ITEM CODE)
            # note that this must be numeric, or else we're not interested
            item_code = sheet.cell_value(r, 1)
            if not isinstance(item_code, (int, float)):
                continue
            row.vendor_code = six.text_type(int(item_code))

            # column C (UNIT UPC)
            # note that this must be present, or else we're not interested
            unit_upc = sheet.cell_value(r, 2)
            if isinstance(unit_upc, float):
                unit_upc = six.text_type(int(unit_upc))
            unit_upc = unit_upc.replace(' ', '')
            if not unit_upc:
                continue
            # note that if UPC is "NA" then we'll still keep the record, it
            # just won't have a UPC value
            if unit_upc != 'NA':
                assert len(unit_upc) == 12
                row.item_entry = unit_upc
                row.upc = GPC(unit_upc, calc_check_digit=False)

            # column E (UNIT PRICE)
            unit_price = sheet.cell_value(r, 4)
            if isinstance(unit_price, float):
                row.unit_cost = decimal.Decimal(six.text_type(unit_price))\
                                       .quantize(decimal.Decimal('0.12345'))
            else:
                match = unit_price_pattern.match(unit_price)
                row.unit_cost = decimal.Decimal(match.group(1))

            # column F (CASE SIZE)
            case_size = sheet.cell_value(r, 5)
            row.case_size = self.parse_case_size(case_size, department_name)
            if department_name == 'Grocery':
                row.size = self.parse_size(case_size)

            # column G (PRICE/CASE)
            price_per_case = sheet.cell_value(r, 6)
            row.case_cost = decimal.Decimal(six.text_type(price_per_case))\
                                   .quantize(decimal.Decimal('0.12345'))

            yield row

    def parse_case_size(self, case_size, department_name):

        # e.g. '2 x 5lb'
        match = re.match(r'^(\d+) x (\d+)lbs?$', case_size)
        if match:
            return int(match.group(1)) * int(match.group(2))

        # e.g. '25 lbs'
        match = re.match(r'^(\d+(?:\.\d+)?) lbs$', case_size)
        if match:
            return decimal.Decimal(match.group(1))

        # e.g. '3 x 5kgs'
        match = re.match(r'^(\d+) x (\d+)kgs$', case_size)
        if match:
            # TODO: assuming 2.2 lb/kg here; what else can we do?
            return 2.2 * int(match.group(1)) * int(match.group(2))

        # e.g. '888 pieces'
        match = re.match(r'^(\d+) pieces$', case_size)
        if match:
            return int(match.group(1))

        # e.g. '12 bars'
        match = re.match(r'^(\d+) bars$', case_size)
        if match:
            return int(match.group(1))

        # e.g. '6 boxes'
        match = re.match(r'^(\d+) boxes$', case_size)
        if match:
            return int(match.group(1))

        if department_name == 'Grocery':

            # e.g. '6 x 8 oz'
            match = re.match(r'^(\d+) x \d+ oz$', case_size)
            if match:
                return int(match.group(1))

            # e.g. '6 x 500 ml'
            match = re.match(r'^(\d+) x \d+ ml$', case_size)
            if match:
                return int(match.group(1))

        raise NotImplementedError("do not know how to parse case size: {}".format(case_size))

    def parse_size(self, case_size):

        # e.g. '6 x 8 oz'
        match = re.match(r'^\d+ x (\d+ oz)$', case_size)
        if match:
            return match.group(1)

        # e.g. '6 x 500 ml'
        match = re.match(r'^\d+ x (\d+ ml)$', case_size)
        if match:
            return match.group(1)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands related to purging of old data
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import shutil
import logging

from rattail.commands.core import Subcommand, date_argument
from rattail.core import Object
from rattail.time import localtime, make_utc


log = logging.getLogger(__name__)


class PurgeSubcommand(Subcommand):
    """
    Base class for subcommands which purge old data
    """
    purge_title = "Object"
    default_before_days = 90

    @property
    def purge_title_plural(self):
        return "{}s".format(self.purge_title)

    def add_parser_args(self, parser):
        super(PurgeSubcommand, self).add_parser_args(parser)

        parser.add_argument('--before', type=date_argument, metavar='DATE',
                            help="Use this date as cutoff, i.e. purge all data "
                            "*before* this date.  If not specified, will use "
                            "--before-days to calculate instead.")

        parser.add_argument('--before-days', type=int, metavar='DAYS',
                            default=self.default_before_days,
                            help="Calculate the cutoff date by subtracting this "
                            "number of days from the current date, i.e. purge all "
                            "data *before* the resulting date.  Note that if you "
                            "specify --before then that date will be used instead "
                            "of calculating one from --before-days.  If neither is "
                            "specified then --before-days is used, with its default "
                            "value of 90 days.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging "
                            "etc. to occur, but rollback (abort) the transaction "
                            "at the end.")

    def run(self, args):
        log.info("will purge things of type: %s", self.purge_title)

        if args.before and args.before_days != self.default_before_days:
            log.warning("specifying both --before and --before-days is "
                        "redundant; --before will take precedence.")

        session = self.make_session()

        # calculate our cutoff date
        if args.before:
            cutoff = args.before
        else:
            today = localtime(self.config).date()
            cutoff = today - datetime.timedelta(days=args.before_days)
        cutoff = datetime.datetime.combine(cutoff, datetime.time(0))
        cutoff = localtime(self.config, cutoff)
        log.info("using %s as cutoff date", cutoff.date())

        # find things, and purge them
        things = self.find_things_to_purge(session, cutoff,
                                           dry_run=args.dry_run)
        log.info("found %s thing(s) to purge", len(things or []))
        if things:
            purged = self.purge_things(session, things, cutoff,
                                       dry_run=args.dry_run)
            log.info("%spurged %s %s",
                     "(would have) " if args.dry_run else "",
                     purged, self.purge_title_plural)

        self.finalize_session(session, dry_run=args.dry_run)
        session.close()

    def find_things_to_purge(self, session, cutoff, dry_run=False):
        pass

    def purge_things(self, session, things, cutoff, dry_run=False):
        result = Object(purged=0)

        def purge(thing, i):
            if self.purge_thing(session, thing, cutoff,
                                dry_run=dry_run):
                result.purged += 1
            if i % 200 == 0:
                session.flush()

        self.progress_loop(purge, things,
                           message="Purging {}".format(self.purge_title_plural))
        return result.purged

    def purge_thing(self, session, thing, cutoff, dry_run=False):
        """
        This method should contain logic which actually "purges" something.
        """
        log.info("purging object: %s", thing)
        return True


class PurgeExport(PurgeSubcommand):
    """
    Base classes for data which purge "export" objects
    """

    @property
    def purge_model_class(self):
        raise NotImplementedError("You must define %s.purge_model_class",
                                  self.__class__.__name__)

    @property
    def purge_title(self):
        return self.purge_model_class.get_model_title()

    @property
    def purge_title_plural(self):
        return self.purge_model_class.get_model_title_plural()

    def find_things_to_purge(self, session, cutoff, dry_run=False):
        model = self.model
        exports = session.query(self.purge_model_class)\
                         .filter(self.purge_model_class.created < make_utc(cutoff))\
                         .all()
        return exports

    def purge_thing(self, session, export, cutoff, dry_run=False):
        """
        This method does the basics, i.e. deleting the export record from the
        database session, and (if not dry-run) deleting any associated files
        from disk.
        """
        uuid = export.uuid
        log.debug("purging export object %s: %s", uuid, export)
        session.delete(export)

        # maybe delete associated files
        if not dry_run:
            session.flush()
            key = self.purge_model_class.export_key
            path = self.config.export_filepath(key, uuid)
            if os.path.exists(path):
                shutil.rmtree(path)

        return True



# -*- coding: utf-8; -*-
"""add product.suggested_price

Revision ID: 11a9145414a9
Revises: c2085638f487
Create Date: 2018-11-08 08:36:27.171039

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '11a9145414a9'
down_revision = 'c2085638f487'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.add_column('product', sa.Column('suggested_price_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('product_fk_suggested_price', 'product', 'product_price', ['suggested_price_uuid'], ['uuid'], use_alter=True)
    op.add_column('product_version', sa.Column('suggested_price_uuid', sa.String(length=32), autoincrement=False, nullable=True))


def downgrade():

    # product
    op.drop_column('product_version', 'suggested_price_uuid')
    op.drop_constraint('product_fk_suggested_price', 'product', type_='foreignkey')
    op.drop_column('product', 'suggested_price_uuid')



# -*- coding: utf-8 -*-
"""add deposit_link

Revision ID: 2d046576154c
Revises: 3f5be78f0f18
Create Date: 2015-02-25 01:13:20.181621

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '2d046576154c'
down_revision = u'3f5be78f0f18'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # deposit_link
    op.create_table('deposit_link',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('code', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('amount', sa.Numeric(precision=5, scale=2), nullable=False),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('code', name=u'deposit_link_uq_code')
                    )
    op.create_table('deposit_link_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('amount', sa.Numeric(precision=5, scale=2), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_deposit_link_version_end_transaction_id'), 'deposit_link_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_deposit_link_version_operation_type'), 'deposit_link_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_deposit_link_version_transaction_id'), 'deposit_link_version', ['transaction_id'], unique=False)

    # product.deposit_link
    op.add_column(u'product', sa.Column('deposit_link_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'product_fk_deposit_link', 'product', 'deposit_link', ['deposit_link_uuid'], ['uuid'])
    op.add_column(u'product_version', sa.Column('deposit_link_uuid', sa.String(length=32), autoincrement=False, nullable=True))


def downgrade():

    # product.deposit_link
    op.drop_column(u'product_version', 'deposit_link_uuid')
    op.drop_constraint(u'product_fk_deposit_link', 'product', type_='foreignkey')
    op.drop_column(u'product', 'deposit_link_uuid')

    # deposit_link
    op.drop_index(op.f('ix_deposit_link_version_transaction_id'), table_name='deposit_link_version')
    op.drop_index(op.f('ix_deposit_link_version_operation_type'), table_name='deposit_link_version')
    op.drop_index(op.f('ix_deposit_link_version_end_transaction_id'), table_name='deposit_link_version')
    op.drop_table('deposit_link_version')
    op.drop_table('deposit_link')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler
"""

from __future__ import unicode_literals, absolute_import

import os
# import re
import tempfile

from sqlalchemy import orm

from rattail.util import load_object, pretty_quantity
from rattail.files import temp_path
from rattail.mail import send_email


class AppHandler(object):
    """
    Base class and default implementation for top-level Rattail app handler.

    aka. "the handler to handle all handlers"

    aka. "one handler to bind them all"
    """
    default_autocompleters = {
        'brands': 'rattail.autocomplete.brands:BrandAutocompleter',
        'customers': 'rattail.autocomplete.customers:CustomerAutocompleter',
        'customers.phone': 'rattail.autocomplete.customers:CustomerPhoneAutocompleter',
        'employees': 'rattail.autocomplete.employees:EmployeeAutocompleter',
        'departments': 'rattail.autocomplete.departments:DepartmentAutocompleter',
        'people': 'rattail.autocomplete.people:PersonAutocompleter',
        'people.employees': 'rattail.autocomplete.people:PersonEmployeeAutocompleter',
        'products': 'rattail.autocomplete.products:ProductAutocompleter',
        'products.all': 'rattail.autocomplete.products:ProductAllAutocompleter',
        'vendors': 'rattail.autocomplete.vendors:VendorAutocompleter',
    }

    def __init__(self, config):
        self.config = config

    def get_autocompleter(self, key, **kwargs):
        """
        Returns an Autocompleter instance corresponding to the given
        key.  Config may determine where the Autocompleter class
        lives, or there are some defaults registered within the app
        handler which may be used.
        """
        spec = self.config.get('rattail', 'autocomplete.{}'.format(key))
        if not spec:
            spec = self.default_autocompleters.get(key)
        if spec:
            return load_object(spec)(self.config)

        raise NotImplementedError("cannot locate autocompleter for key: {}".format(key))

    def get_board_handler(self, **kwargs):
        if not hasattr(self, 'board_handler'):
            from rattail.board import get_board_handler
            self.board_handler = get_board_handler(self.config, **kwargs)
        return self.board_handler

    def get_clientele_handler(self, **kwargs):
        if not hasattr(self, 'clientele_handler'):
            from rattail.clientele import get_clientele_handler
            self.clientele_handler = get_clientele_handler(self.config, **kwargs)
        return self.clientele_handler

    def get_employment_handler(self, **kwargs):
        if not hasattr(self, 'employment_handler'):
            from rattail.employment import get_employment_handler
            self.employment_handler = get_employment_handler(self.config, **kwargs)
        return self.employment_handler

    def get_feature_handler(self, **kwargs):
        if not hasattr(self, 'feature_handler'):
            from rattail.features import FeatureHandler
            self.feature_handler = FeatureHandler(self.config, **kwargs)
        return self.feature_handler

    def get_email_handler(self, **kwargs):
        if not hasattr(self, 'email_handler'):
            from rattail.mail import get_email_handler
            self.email_handler = get_email_handler(self.config, **kwargs)
        return self.email_handler

    # TODO: is it helpful to expose this? or more confusing?
    get_mail_handler = get_email_handler

    def get_membership_handler(self, **kwargs):
        """
        Returns a reference to the configured Membership Handler.

        See also :doc:`rattail-manual:base/handlers/other/membership`.
        """
        if not hasattr(self, 'membership_handler'):
            spec = self.config.get('rattail', 'membership.handler',
                                   default='rattail.membership:MembershipHandler')
            factory = load_object(spec)
            self.membership_handler = factory(self.config, **kwargs)
        return self.membership_handler

    def get_people_handler(self, **kwargs):
        """
        Returns a reference to the configured People Handler.

        See also :doc:`rattail-manual:base/handlers/other/people`.
        """
        if not hasattr(self, 'people_handler'):
            spec = self.config.get('rattail', 'people.handler',
                                   default='rattail.people:PeopleHandler')
            factory = load_object(spec)
            self.people_handler = factory(self.config, **kwargs)
        return self.people_handler

    def get_products_handler(self, **kwargs):
        if not hasattr(self, 'products_handler'):
            from rattail.products import get_products_handler
            self.products_handler = get_products_handler(self.config, **kwargs)
        return self.products_handler

    def get_report_handler(self, **kwargs):
        if not hasattr(self, 'report_handler'):
            from rattail.reporting import get_report_handler
            self.report_handler = get_report_handler(self.config, **kwargs)
        return self.report_handler

    def get_session(self, obj):
        """
        Returns the SQLAlchemy session with which the given object is
        associated.  Simple convenience wrapper around
        ``sqlalchemy.orm.object_session()``.
        """
        return orm.object_session(obj)

    def make_session(self, **kwargs):
        """
        Creates and returns a new SQLAlchemy session for the Rattail DB.
        """
        from rattail.db import Session
        return Session(**kwargs)

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes
        :func:`rattail.db.cache.cache_model()` with the given model
        and keyword arguments.
        """
        from rattail.db import cache
        return cache.cache_model(session, model, **kwargs)

    def make_temp_dir(self, **kwargs):
        """
        Create a temporary directory.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkdtemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return tempfile.mkdtemp(**kwargs)

    def make_temp_file(self, **kwargs):
        """
        Reserve a temporary filename.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkstemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return temp_path(**kwargs)

    def normalize_phone_number(self, number):
        """
        Normalize the given phone number, to a "common" format that
        can be more easily worked with for sync logic etc.
        """
        from rattail.db.util import normalize_phone_number

        return normalize_phone_number(number)

    def phone_number_is_invalid(self, number):
        """
        This method should validate the given phone number string, and if the
        number is *not* considered valid, this method should return the reason
        as string.  If the number is valid, returns ``None``.
        """
        # strip non-numeric chars, and make sure we have 10 left
        normal = self.normalize_phone_number(number)
        if len(normal) != 10:
            return "Phone number must have 10 digits"

    def format_phone_number(self, number):
        """
        Returns a "properly formatted" string based on the given phone number.
        """
        from rattail.db.util import format_phone_number

        return format_phone_number(number)

    def render_currency(self, value, scale=2, **kwargs):
        """
        Must return a human-friendly display string for the given currency
        value, e.g. ``Decimal('4.20')`` becomes ``"$4.20"``.
        """
        if value is not None:
            if value < 0:
                fmt = "(${{:0,.{}f}})".format(scale)
                return fmt.format(0 - value)
            fmt = "${{:0,.{}f}}".format(scale)
            return fmt.format(value)

    def render_quantity(self, value, **kwargs):
        """
        Return a human-friendly display string for the given quantity
        value, e.g. ``1.000`` becomes ``"1"``.
        """
        return pretty_quantity(value, **kwargs)

    def render_date(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``date``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d')

    def render_datetime(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``datetime``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d %I:%M:%S %p')

    def send_email(self, key, data={}, **kwargs):
        """
        Send an email message of the given type.

        See :func:`rattail.mail.send_email()` for more info.
        """
        send_email(self.config, key, data, **kwargs)


class GenericHandler(object):
    """
    Base class for misc. "generic" feature handlers.

    Most handlers which exist for sake of business logic, should inherit from
    this.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.enum = self.config.get_enum()
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def get_session(self, obj):
        """
        Convenience wrapper around :meth:`AppHandler.get_session()`.
        """
        return self.app.get_session(obj)

    def make_session(self):
        """
        Convenience wrapper around :meth:`AppHandler.make_session()`.
        """
        return self.app.make_session()

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.
        """
        return self.app.cache_model(session, model, **kwargs)


def make_app(config, **kwargs):
    """
    Create and return the configured :class:`AppHandler` instance.
    """
    spec = config.get('rattail', 'app.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = AppHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for "customer order" batches
"""

from __future__ import unicode_literals, absolute_import, division

import re

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler


class CustomerOrderBatchHandler(BatchHandler):
    """
    Handler for all "customer order" batches, regardless of "mode".  The
    handler must inspect the
    :attr:`~rattail.db.model.batch.custorder.CustomerOrderBatch.mode` attribute
    of each batch it deals with, in order to determine which logic to apply.

    .. attribute:: has_custom_product_autocomplete

       If true, this flag indicates that the handler provides custom
       autocomplete logic for use when selecting a product while
       creating a new order.
    """
    batch_model_class = model.CustomerOrderBatch
    has_custom_product_autocomplete = False
    nondigits_pattern = re.compile(r'\D')

    def init_batch(self, batch, progress=None, **kwargs):
        """
        Assign the "local" store to the batch, if applicable.
        """
        session = self.app.get_session(batch)
        batch.store = self.config.get_store(session)

    def new_order_requires_customer(self):
        """
        Returns a boolean indicating whether a *new* "customer order"
        in fact requires a proper customer account, or not.  Note that
        in all cases a new order requires a *person* to associate
        with, but technically the customer is optional, unless this
        returns true.
        """
        return self.config.getbool('rattail.custorders',
                                   'new_order_requires_customer',
                                   default=False)

    def assign_contact(self, batch, customer=None, person=None, **kwargs):
        """
        Assign the customer and/or person "contact" for the order.
        """
        clientele = self.app.get_clientele_handler()
        customer_required = self.new_order_requires_customer()

        # nb. person is always required
        if customer and not person:
            person = clientele.get_person(customer)
        if not person:
            raise ValueError("Must specify a person")

        # customer may or may not be optional
        if person and not customer:
            customer = clientele.get_customer(person)
        if customer_required and not customer:
            raise ValueError("Must specify a customer account")

        # assign contact
        batch.customer = customer
        batch.person = person

        # update phone/email per new contact
        batch.phone_number = None
        batch.email_address = None
        if customer_required:
            batch.phone_number = clientele.get_first_phone_number(customer)
            batch.email_address = clientele.get_first_email_address(customer)
        else:
            batch.phone_number = person.first_phone_number()
            batch.email_address = person.first_email_address()

        session = self.app.get_session(batch)
        session.flush()
        session.refresh(batch)

    def unassign_contact(self, batch, **kwargs):
        """
        Unassign the customer and/or person "contact" for the order.
        """
        batch.customer = None
        batch.person = None
        batch.phone_number = None
        batch.email_address = None

        session = self.app.get_session(batch)
        session.flush()
        session.refresh(batch)

    def get_case_size_for_product(self, product):
        if product.case_size:
            return product.case_size

        cost = product.cost
        if cost:
            return cost.case_size

    def get_phone_search_term(self, term):
        """
        Try to figure out if the given search term represents a whole
        or partial phone number, and if so return just the digits.
        """
        digits = self.nondigits_pattern.sub('', term)
        if digits and len(digits) >= 4:
            return digits

    def customer_autocomplete(self, session, term, **kwargs):
        """
        Override the Customer autocomplete, to search by phone number
        as well as name.
        """
        model = self.model

        # define the base query
        query = session.query(model.Customer)\
                       .options(orm.joinedload(model.Customer.phones))

        # does search term look like a phone number?
        phone_term = self.get_phone_search_term(term)
        if phone_term:

            # yep, so just search for the phone number
            query = query.join(model.CustomerPhoneNumber,
                               model.CustomerPhoneNumber.parent_uuid == model.Customer.uuid)
            query = query.filter(sa.func.regexp_replace(model.CustomerPhoneNumber.number,
                                                        r'\D', '', 'g')\
                                 .like('%{}%'.format(phone_term)))

        else: # term does not look like a phone number

            # so just search by name
            criteria = [model.Customer.name.ilike('%{}%'.format(word))
                        for word in term.split()]
            query = query.filter(sa.and_(*criteria))

        # oh, and sort by something useful
        query = query.order_by(model.Customer.name)

        # generate result list from query
        results = []
        for customer in query:
            phone = customer.first_phone()
            if phone:
                label = "{} {}".format(customer.name, phone.number)
            else:
                label = customer.name
            results.append({'value': customer.uuid,
                            'label': label,
                            'display': customer.name})

        return results

    def person_autocomplete(self, session, term, **kwargs):
        """
        Override the Person autocomplete, to search by phone number as
        well as name.
        """
        model = self.model

        # define the base query
        query = session.query(model.Person)\
                       .options(orm.joinedload(model.Person.phones))

        # does search term look like a phone number?
        phone_term = self.get_phone_search_term(term)
        if phone_term:

            # yep, so just search for the phone number
            query = query.join(model.PersonPhoneNumber,
                               model.PersonPhoneNumber.parent_uuid == model.Person.uuid)
            query = query.filter(sa.func.regexp_replace(model.PersonPhoneNumber.number,
                                                        r'\D', '', 'g')\
                                 .like('%{}%'.format(phone_term)))

        else: # term does not look like a phone number

            # so just search by name
            criteria = [model.Person.display_name.ilike('%{}%'.format(word))
                        for word in term.split()]
            query = query.filter(sa.and_(*criteria))

        # oh, and sort by something useful
        query = query.order_by(model.Person.display_name)

        # generate result list from query
        results = []
        for person in query:
            phone = person.first_phone()
            if phone:
                label = "{} {}".format(person.display_name, phone.number)
            else:
                label = person.display_name
            results.append({'value': person.uuid,
                            'label': label,
                            'display': person.display_name})

        return results

    def get_customer_info(self, batch, **kwargs):
        """
        Return a data dict containing misc. info pertaining to the
        customer/person for the order batch.
        """
        info = {
            'customer_uuid': None,
            'person_uuid': None,
            'phone_number': None,
            'email_address': None,
        }

        if batch.customer:
            info['customer_uuid'] = batch.customer.uuid
            phone = batch.customer.first_phone()
            if phone:
                info['phone_number'] = phone.number
            email = batch.customer.first_email()
            if email:
                info['email_address'] = email.address

        if batch.person:
            info['person_uuid'] = batch.person.uuid
            if not info['phone_number']:
                phone = batch.person.first_phone()
                if phone:
                    info['phone_number'] = phone.number
                email = batch.person.first_email()
                if email:
                    info['email_address'] = email.address

        return info

    def custom_product_autocomplete(self, session, term, **kwargs):
        """
        For the given term, this should return a (possibly empty) list
        of products which "match" the term.  Each element in the list
        should be a dict with "label" and "value" keys.
        """
        raise NotImplementedError("Please define the "
                                  "{}.custom_product_autocomplete() "
                                  "method.".format(__class__.__name__))

    def refresh_row(self, row):
        if not row.product:
            if row.item_entry:
                session = orm.object_session(row)
                # TODO: should do more than just query for uuid here
                product = session.query(model.Product).get(row.item_entry)
                if product:
                    row.product = product
            if not row.product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        product = row.product
        row.product_upc = product.upc
        row.product_brand = six.text_type(product.brand or "")
        row.product_description = product.description
        row.product_size = product.size
        row.product_weighed = product.weighed
        row.case_quantity = self.get_case_size_for_product(product)

        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None

        cost = product.cost
        row.product_unit_cost = cost.unit_cost if cost else None

        regprice = product.regular_price
        row.unit_price = regprice.price if regprice else None

        # we need to know if total price is updated
        old_total = row.total_price

        # maybe update total price
        if row.unit_price is None:
            row.total_price = None
        elif not row.unit_price:
            row.total_price = 0
        else:
            row.total_price = row.unit_price * row.order_quantity
            if row.order_uom == self.enum.UNIT_OF_MEASURE_CASE:
                row.total_price *= (row.case_quantity or 1)

        # update total price for batch too, if it changed
        if row.total_price != old_total:
            batch = row.batch
            batch.total_price = ((batch.total_price or 0)
                                 + (row.total_price or 0)
                                 - (old_total or 0))

        row.status_code = row.STATUS_OK

    def remove_row(self, row):
        batch = row.batch

        if not row.removed:
            row.removed = True

            if row.total_price:
                batch.total_price = (batch.total_price or 0) - row.total_price

        self.refresh_batch_status(batch)

    def execute(self, batch, user=None, progress=None, **kwargs):
        """
        Default behavior here will simply create a new (proper) Customer Order
        based on the batch contents.  Override as needed.
        """
        batch_fields = [
            'store',
            'id',
            'customer',
            'person',
            'phone_number',
            'email_address',
            'total_price',
        ]

        order = model.CustomerOrder()
        order.created_by = user
        order.status_code = self.enum.CUSTORDER_STATUS_ORDERED
        for field in batch_fields:
            setattr(order, field, getattr(batch, field))

        row_fields = [
            'product',
            'product_upc',
            'product_brand',
            'product_description',
            'product_size',
            'product_weighed',
            'department_number',
            'department_name',
            'case_quantity',
            'order_quantity',
            'order_uom',
            'product_unit_cost',
            'unit_price',
            'discount_percent',
            'total_price',
            'paid_amount',
            'payment_transaction_number',
        ]

        def convert(row, i):
            item = model.CustomerOrderItem()
            item.sequence = i
            item.status_code = self.enum.CUSTORDER_ITEM_STATUS_INITIATED
            for field in row_fields:
                setattr(item, field, getattr(row, field))
            order.items.append(item)

            # attach event
            item.events.append(model.CustomerOrderItemEvent(
                type_code=self.enum.CUSTORDER_ITEM_EVENT_INITIATED,
                user=user))

        self.progress_loop(convert, batch.active_rows(), progress,
                           message="Converting batch rows to order items")

        session = orm.object_session(batch)
        session.add(order)
        session.flush()

        return order



# -*- coding: utf-8 -*-
"""add Member.number

Revision ID: be2cb8322066
Revises: dfc1ed686f3f
Create Date: 2020-03-18 12:25:28.386086

"""

# revision identifiers, used by Alembic.
revision = 'be2cb8322066'
down_revision = 'dfc1ed686f3f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member
    op.add_column('member', sa.Column('number', sa.Integer(), nullable=True))
    op.add_column('member_version', sa.Column('number', sa.Integer(), autoincrement=False, nullable=True))


def downgrade():

    # member
    op.drop_column('member_version', 'number')
    op.drop_column('member', 'number')



# -*- coding: utf-8 -*-
"""add some product flags

Revision ID: 7ba105af22a7
Revises: 9305c1f493b4
Create Date: 2017-02-22 19:05:30.964165

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '7ba105af22a7'
down_revision = u'9305c1f493b4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product
    op.add_column('product', sa.Column('gluten_free', sa.Boolean(), nullable=True))
    op.add_column('product', sa.Column('kosher', sa.Boolean(), nullable=True))
    op.add_column('product', sa.Column('sugar_free', sa.Boolean(), nullable=True))
    op.add_column('product', sa.Column('vegan', sa.Boolean(), nullable=True))
    op.add_column('product', sa.Column('vegetarian', sa.Boolean(), nullable=True))


def downgrade():

    # product
    op.drop_column('product', 'vegetarian')
    op.drop_column('product', 'vegan')
    op.drop_column('product', 'sugar_free')
    op.drop_column('product', 'kosher')
    op.drop_column('product', 'gluten_free')



# -*- coding: utf-8; -*-
"""add pricing_batch.true_unit_cost

Revision ID: c2085638f487
Revises: 35b451f8cc27
Create Date: 2018-11-02 18:42:35.715199

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c2085638f487'
down_revision = '35b451f8cc27'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'price_margin', type_=sa.Numeric(precision=6, scale=3))
    op.add_column('batch_pricing_row', sa.Column('true_margin', sa.Numeric(precision=6, scale=3), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('true_unit_cost', sa.Numeric(precision=9, scale=5), nullable=True))


def downgrade():

    # batch_pricing_row
    op.drop_column('batch_pricing_row', 'true_unit_cost')
    op.drop_column('batch_pricing_row', 'true_margin')
    op.alter_column('batch_pricing_row', 'price_margin', type_=sa.Numeric(precision=5, scale=3))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete Handler for Products
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.autocomplete import Autocompleter
from rattail.db import model


class ProductAutocompleter(Autocompleter):
    """
    Autocompleter for Products
    """
    autocompleter_key = 'products'
    model_class = model.Product
    autocomplete_fieldname = 'description'

    def make_autocomplete_query(self, session, term, **kwargs):
        model = self.model
        query = session.query(model.Product)\
                       .outerjoin(model.Brand)\
                       .filter(sa.or_(
                           model.Brand.name.ilike('%{}%'.format(term)),
                           model.Product.description.ilike('%{}%'.format(term))))

        query = self.filter_autocomplete_query(session, query)

        query = query.order_by(model.Brand.name, 
                               model.Product.description)\
                     .options(orm.joinedload(model.Product.brand))
        return query

    def filter_autocomplete_query(self, session, query, **kwargs):
        # do not show "deleted" items by default
        query = query.filter(model.Product.deleted == False)
        return query

    def autocomplete_display(self, product):
        return product.full_description


class ProductAllAutocompleter(ProductAutocompleter):
    """
    Autocompleter for Products, which shows *all* results, including
    "deleted" items etc.
    """

    def filter_autocomplete_query(self, session, query, **kwargs):
        return query



# -*- coding: utf-8 -*-
"""add product versioning

Revision ID: a28c3583c8f0
Revises: aa5962fb998e
Create Date: 2017-07-14 22:31:25.558644

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a28c3583c8f0'
down_revision = u'aa5962fb998e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # brand_version
    op.create_table('brand_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_brand_version_end_transaction_id'), 'brand_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_brand_version_operation_type'), 'brand_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_brand_version_transaction_id'), 'brand_version', ['transaction_id'], unique=False)

    # category_version
    op.create_table('category_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=8), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_category_version_end_transaction_id'), 'category_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_category_version_operation_type'), 'category_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_category_version_transaction_id'), 'category_version', ['transaction_id'], unique=False)

    # department_version
    op.create_table('department_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_department_version_end_transaction_id'), 'department_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_department_version_operation_type'), 'department_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_department_version_transaction_id'), 'department_version', ['transaction_id'], unique=False)

    # deposit_link_version
    op.create_table('deposit_link_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('amount', sa.Numeric(precision=5, scale=2), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_deposit_link_version_end_transaction_id'), 'deposit_link_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_deposit_link_version_operation_type'), 'deposit_link_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_deposit_link_version_transaction_id'), 'deposit_link_version', ['transaction_id'], unique=False)

    # employee_shift_worked_version
    op.create_table('employee_shift_worked_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('store_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('punch_in', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('punch_out', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_employee_shift_worked_version_end_transaction_id'), 'employee_shift_worked_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_shift_worked_version_operation_type'), 'employee_shift_worked_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_shift_worked_version_transaction_id'), 'employee_shift_worked_version', ['transaction_id'], unique=False)

    # employee_x_department_version
    op.create_table('employee_x_department_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_employee_x_department_version_end_transaction_id'), 'employee_x_department_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_x_department_version_operation_type'), 'employee_x_department_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_x_department_version_transaction_id'), 'employee_x_department_version', ['transaction_id'], unique=False)

    # employee_x_store_version
    op.create_table('employee_x_store_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('store_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_employee_x_store_version_end_transaction_id'), 'employee_x_store_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_x_store_version_operation_type'), 'employee_x_store_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_x_store_version_transaction_id'), 'employee_x_store_version', ['transaction_id'], unique=False)

    # family_version
    op.create_table('family_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_family_version_end_transaction_id'), 'family_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_family_version_operation_type'), 'family_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_family_version_transaction_id'), 'family_version', ['transaction_id'], unique=False)

    # label_profile_version
    op.create_table('label_profile_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=3), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('printer_spec', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('formatter_spec', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('format', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('visible', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_label_profile_version_end_transaction_id'), 'label_profile_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_label_profile_version_operation_type'), 'label_profile_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_label_profile_version_transaction_id'), 'label_profile_version', ['transaction_id'], unique=False)

    # product_code_version
    op.create_table('product_code_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_product_code_version_end_transaction_id'), 'product_code_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_code_version_operation_type'), 'product_code_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_code_version_transaction_id'), 'product_code_version', ['transaction_id'], unique=False)

    # product_cost_version
    op.create_table('product_cost_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('vendor_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('case_size', sa.Numeric(precision=9, scale=4), autoincrement=False, nullable=True),
                    sa.Column('case_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('pack_size', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('pack_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('unit_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('effective', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('discontinued', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_product_cost_version_end_transaction_id'), 'product_cost_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_cost_version_operation_type'), 'product_cost_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_cost_version_transaction_id'), 'product_cost_version', ['transaction_id'], unique=False)

    # product_price_version
    op.create_table('product_price_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('type', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('level', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('starts', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('ends', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('price', sa.Numeric(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('multiple', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('pack_price', sa.Numeric(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('pack_multiple', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_product_price_version_end_transaction_id'), 'product_price_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_price_version_operation_type'), 'product_price_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_price_version_transaction_id'), 'product_price_version', ['transaction_id'], unique=False)

    # product_version
    op.create_table('product_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), autoincrement=False, nullable=True),
                    sa.Column('scancode', sa.String(length=14), autoincrement=False, nullable=True),
                    sa.Column('item_id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('item_type', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('subdepartment_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('category_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('family_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('report_code_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('deposit_link_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('tax_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('brand_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('description2', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('size', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('unit_size', sa.Numeric(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('unit_of_measure', sa.String(length=4), autoincrement=False, nullable=True),
                    sa.Column('uom_abbreviation', sa.String(length=4), autoincrement=False, nullable=True),
                    sa.Column('weighed', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('unit_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('pack_size', sa.Numeric(precision=9, scale=4), autoincrement=False, nullable=True),
                    sa.Column('case_size', sa.Numeric(precision=9, scale=4), autoincrement=False, nullable=True),
                    sa.Column('organic', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('kosher', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('vegan', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('vegetarian', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('gluten_free', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('sugar_free', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('not_for_sale', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('discontinued', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('deleted', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('regular_price_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('current_price_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('discountable', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('special_order', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('food_stampable', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('tax1', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('tax2', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('tax3', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('ingredients', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('notes', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_product_version_end_transaction_id'), 'product_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_version_operation_type'), 'product_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_version_transaction_id'), 'product_version', ['transaction_id'], unique=False)

    # report_code_version
    op.create_table('report_code_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_report_code_version_end_transaction_id'), 'report_code_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_report_code_version_operation_type'), 'report_code_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_report_code_version_transaction_id'), 'report_code_version', ['transaction_id'], unique=False)

    # role_version
    op.create_table('role_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.String(length=25), autoincrement=False, nullable=True),
                    sa.Column('session_timeout', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_role_version_end_transaction_id'), 'role_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_role_version_operation_type'), 'role_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_role_version_transaction_id'), 'role_version', ['transaction_id'], unique=False)

    # subdepartment_version
    op.create_table('subdepartment_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_subdepartment_version_end_transaction_id'), 'subdepartment_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_subdepartment_version_operation_type'), 'subdepartment_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_subdepartment_version_transaction_id'), 'subdepartment_version', ['transaction_id'], unique=False)

    # tax_version
    op.create_table('tax_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('rate', sa.Numeric(precision=7, scale=5), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_tax_version_end_transaction_id'), 'tax_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_tax_version_operation_type'), 'tax_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_tax_version_transaction_id'), 'tax_version', ['transaction_id'], unique=False)

    # user_version
    op.create_table('user_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('username', sa.String(length=25), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('active', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('active_sticky', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_user_version_end_transaction_id'), 'user_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_user_version_operation_type'), 'user_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_user_version_transaction_id'), 'user_version', ['transaction_id'], unique=False)

    # user_x_role_version
    op.create_table('user_x_role_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('role_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_user_x_role_version_end_transaction_id'), 'user_x_role_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_user_x_role_version_operation_type'), 'user_x_role_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_user_x_role_version_transaction_id'), 'user_x_role_version', ['transaction_id'], unique=False)


def downgrade():

    # user_x_role_version
    op.drop_index(op.f('ix_user_x_role_version_transaction_id'), table_name='user_x_role_version')
    op.drop_index(op.f('ix_user_x_role_version_operation_type'), table_name='user_x_role_version')
    op.drop_index(op.f('ix_user_x_role_version_end_transaction_id'), table_name='user_x_role_version')
    op.drop_table('user_x_role_version')

    # user_version
    op.drop_index(op.f('ix_user_version_transaction_id'), table_name='user_version')
    op.drop_index(op.f('ix_user_version_operation_type'), table_name='user_version')
    op.drop_index(op.f('ix_user_version_end_transaction_id'), table_name='user_version')
    op.drop_table('user_version')

    # tax_version
    op.drop_index(op.f('ix_tax_version_transaction_id'), table_name='tax_version')
    op.drop_index(op.f('ix_tax_version_operation_type'), table_name='tax_version')
    op.drop_index(op.f('ix_tax_version_end_transaction_id'), table_name='tax_version')
    op.drop_table('tax_version')

    # subdepartment_version
    op.drop_index(op.f('ix_subdepartment_version_transaction_id'), table_name='subdepartment_version')
    op.drop_index(op.f('ix_subdepartment_version_operation_type'), table_name='subdepartment_version')
    op.drop_index(op.f('ix_subdepartment_version_end_transaction_id'), table_name='subdepartment_version')
    op.drop_table('subdepartment_version')

    # role_version
    op.drop_index(op.f('ix_role_version_transaction_id'), table_name='role_version')
    op.drop_index(op.f('ix_role_version_operation_type'), table_name='role_version')
    op.drop_index(op.f('ix_role_version_end_transaction_id'), table_name='role_version')
    op.drop_table('role_version')

    # report_code_version
    op.drop_index(op.f('ix_report_code_version_transaction_id'), table_name='report_code_version')
    op.drop_index(op.f('ix_report_code_version_operation_type'), table_name='report_code_version')
    op.drop_index(op.f('ix_report_code_version_end_transaction_id'), table_name='report_code_version')
    op.drop_table('report_code_version')

    # product_version
    op.drop_index(op.f('ix_product_version_transaction_id'), table_name='product_version')
    op.drop_index(op.f('ix_product_version_operation_type'), table_name='product_version')
    op.drop_index(op.f('ix_product_version_end_transaction_id'), table_name='product_version')
    op.drop_table('product_version')

    # product_price_version
    op.drop_index(op.f('ix_product_price_version_transaction_id'), table_name='product_price_version')
    op.drop_index(op.f('ix_product_price_version_operation_type'), table_name='product_price_version')
    op.drop_index(op.f('ix_product_price_version_end_transaction_id'), table_name='product_price_version')
    op.drop_table('product_price_version')

    # product_cost_version
    op.drop_index(op.f('ix_product_cost_version_transaction_id'), table_name='product_cost_version')
    op.drop_index(op.f('ix_product_cost_version_operation_type'), table_name='product_cost_version')
    op.drop_index(op.f('ix_product_cost_version_end_transaction_id'), table_name='product_cost_version')
    op.drop_table('product_cost_version')

    # product_code_version
    op.drop_index(op.f('ix_product_code_version_transaction_id'), table_name='product_code_version')
    op.drop_index(op.f('ix_product_code_version_operation_type'), table_name='product_code_version')
    op.drop_index(op.f('ix_product_code_version_end_transaction_id'), table_name='product_code_version')
    op.drop_table('product_code_version')

    # label_profile_version
    op.drop_index(op.f('ix_label_profile_version_transaction_id'), table_name='label_profile_version')
    op.drop_index(op.f('ix_label_profile_version_operation_type'), table_name='label_profile_version')
    op.drop_index(op.f('ix_label_profile_version_end_transaction_id'), table_name='label_profile_version')
    op.drop_table('label_profile_version')

    # family_version
    op.drop_index(op.f('ix_family_version_transaction_id'), table_name='family_version')
    op.drop_index(op.f('ix_family_version_operation_type'), table_name='family_version')
    op.drop_index(op.f('ix_family_version_end_transaction_id'), table_name='family_version')
    op.drop_table('family_version')

    # employee_x_store_version
    op.drop_index(op.f('ix_employee_x_store_version_transaction_id'), table_name='employee_x_store_version')
    op.drop_index(op.f('ix_employee_x_store_version_operation_type'), table_name='employee_x_store_version')
    op.drop_index(op.f('ix_employee_x_store_version_end_transaction_id'), table_name='employee_x_store_version')
    op.drop_table('employee_x_store_version')

    # employee_x_department_version
    op.drop_index(op.f('ix_employee_x_department_version_transaction_id'), table_name='employee_x_department_version')
    op.drop_index(op.f('ix_employee_x_department_version_operation_type'), table_name='employee_x_department_version')
    op.drop_index(op.f('ix_employee_x_department_version_end_transaction_id'), table_name='employee_x_department_version')
    op.drop_table('employee_x_department_version')

    # employee_shift_worked_version
    op.drop_index(op.f('ix_employee_shift_worked_version_transaction_id'), table_name='employee_shift_worked_version')
    op.drop_index(op.f('ix_employee_shift_worked_version_operation_type'), table_name='employee_shift_worked_version')
    op.drop_index(op.f('ix_employee_shift_worked_version_end_transaction_id'), table_name='employee_shift_worked_version')
    op.drop_table('employee_shift_worked_version')

    # deposit_link_version
    op.drop_index(op.f('ix_deposit_link_version_transaction_id'), table_name='deposit_link_version')
    op.drop_index(op.f('ix_deposit_link_version_operation_type'), table_name='deposit_link_version')
    op.drop_index(op.f('ix_deposit_link_version_end_transaction_id'), table_name='deposit_link_version')
    op.drop_table('deposit_link_version')

    # department_version
    op.drop_index(op.f('ix_department_version_transaction_id'), table_name='department_version')
    op.drop_index(op.f('ix_department_version_operation_type'), table_name='department_version')
    op.drop_index(op.f('ix_department_version_end_transaction_id'), table_name='department_version')
    op.drop_table('department_version')

    # category_version
    op.drop_index(op.f('ix_category_version_transaction_id'), table_name='category_version')
    op.drop_index(op.f('ix_category_version_operation_type'), table_name='category_version')
    op.drop_index(op.f('ix_category_version_end_transaction_id'), table_name='category_version')
    op.drop_table('category_version')

    # brand_version
    op.drop_index(op.f('ix_brand_version_transaction_id'), table_name='brand_version')
    op.drop_index(op.f('ix_brand_version_operation_type'), table_name='brand_version')
    op.drop_index(op.f('ix_brand_version_end_transaction_id'), table_name='brand_version')
    op.drop_table('brand_version')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Changes Interface
"""

from __future__ import unicode_literals, absolute_import

import logging

from packaging.version import parse as parse_version
import sqlalchemy as sa
from sqlalchemy.orm import object_mapper, RelationshipProperty
from sqlalchemy.orm.interfaces import SessionExtension
from sqlalchemy.orm.session import Session

try:
    from sqlalchemy.event import listen
except ImportError:
    listen = None # old SQLAlchemy; will have to work around that, below

from rattail.db import model
from rattail.core import get_uuid
from rattail.util import load_object

try:
    from rattail.db.continuum import versioning_manager
except ImportError:             # assume no continuum
    versioning_manager = None


log = logging.getLogger(__name__)


def record_changes(session, recorder=None, config=None):
    """
    Record all relevant data changes which occur within a session.

    :param session: A :class:`sqlalchemy:sqlalchemy.orm.session.Session` class,
       or instance thereof.
    """
    if isinstance(recorder, ChangeRecorder):
        pass
    elif callable(recorder):
        recorder = recorder()
    elif recorder is None:
        if config:
            spec = config.get('rattail.db', 'changes.recorder', usedb=False)
            if spec:
                recorder = load_object(spec)(config)
        if not recorder:
            recorder = ChangeRecorder(config)
    else:
        raise ValueError("Invalid 'recorder' parameter: {}".format(repr(recorder)))

    if listen:
        listen(session, 'before_flush', recorder)
    else:
        extension = ChangeRecorderExtension(recorder)
        if isinstance(session, Session):
            session.extensions.append(extension)
        else:
            session.configure(extension=extension)

    session.rattail_record_changes = True
    session.rattail_change_recorder = recorder


class ChangeRecorder(object):
    """
    Listener for session ``before_flush`` events.

    This class is responsible for adding stub records to the ``changes`` table,
    which will in turn be used by the database synchronizer to manage change
    data propagation.
    """
    ignored_classes = (
        model.Setting,
        model.Change,
        model.DataSyncChange,
    )

    # once upon a time we supposedly needed to specify `passive=True` when
    # invoking `session.is_modified()` but that has apparently been deprecated
    # for some time now.  we likely should just require SA>=0.8 instead of
    # maintaining this logic?
    is_modified_kw = {}
    if parse_version(sa.__version__) < parse_version('0.8'):
        is_modified_kw['passive'] = True

    def __init__(self, config):
        self.config = config
        self.model = config.get_model()

    def __call__(self, session, flush_context, instances):
        """
        Method invoked when session ``before_flush`` event occurs.
        """
        # TODO: what a mess, need to look into this again at some point...
        # # TODO: Not sure if our event replaces the one registered by Continuum,
        # # or what.  But this appears to be necessary to keep that system
        # # working when we enable ours...
        # if versioning_manager:
        #     versioning_manager.before_flush(session, flush_context, instances)

        for obj in session.deleted:
            if not self.ignore_object(obj):
                self.process_deleted_object(session, obj)
        for obj in session.new:
            if not self.ignore_object(obj):
                self.process_new_object(session, obj)
        for obj in session.dirty:
            if not self.ignore_object(obj) and session.is_modified(obj, **self.is_modified_kw):
                # Orphaned objects which really are pending deletion show up in
                # session.dirty instead of session.deleted, hence this check.
                # https://groups.google.com/d/msg/sqlalchemy/H4nQTHphc0M/Xr8-Cgra0Z4J
                if self.is_deletable_orphan(obj):
                    self.process_deleted_object(session, obj)
                else:
                    self.process_dirty_object(session, obj)

    def ignore_object(self, obj):
        """
        Return ``True`` if changes for the given object should be ignored.
        """
        # ignore certain classes per declaration
        if isinstance(obj, self.ignored_classes):
            return True

        # TODO: is there a smarter way to check?
        # definitely don't care about changes to any version tables
        if obj.__class__.__name__.endswith('Version'):
            return True

        # ignore LabelProfile objects which are *not* marked "sync me"
        if isinstance(obj, model.LabelProfile) and not obj.sync_me:
            return True

        return False # i.e. don't ignore

    def process_new_object(self, session, obj):
        """
        Record changes as appropriate, for the given 'new' object.
        """
        self.record_rattail_change(session, obj, type_='new')

    def process_dirty_object(self, session, obj):
        """
        Record changes as appropriate, for the given 'dirty' object.
        """
        self.record_rattail_change(session, obj, type_='dirty')

    def process_deleted_object(self, session, obj):
        """
        Record changes as appropriate, for the given 'deleted' object.
        """
        # TODO: should perhaps find a "cleaner" way to handle these..?

        # mark Person as dirty, when contact info is removed
        if isinstance(obj, (model.PersonEmailAddress, model.PersonPhoneNumber, model.PersonMailingAddress)):
            self.record_rattail_change(session, obj.person, type_='dirty')

        # mark Employee as dirty, when department info is removed
        if isinstance(obj, (model.EmployeeStore, model.EmployeeDepartment)):
            self.record_rattail_change(session, obj.employee, type_='dirty')

        # mark Product as dirty, when cost info is removed
        if isinstance(obj, model.ProductCost):
            self.record_rattail_change(session, obj.product, type_='dirty')

        self.record_rattail_change(session, obj, type_='deleted')

    def is_deletable_orphan(self, instance):
        """
        Determine if an object is an orphan and pending deletion.
        """
        mapper = object_mapper(instance)
        for property_ in mapper.iterate_properties:
            if isinstance(property_, RelationshipProperty):
                relationship = property_

                # Does this relationship refer back to the instance class?
                backref = relationship.backref or relationship.back_populates
                if backref:

                    # Does the other class mapper's relationship wish to delete orphans?
                    # other_relationship = relationship.mapper.relationships[backref]

                    # Sometimes backrefs are tuples; first element is name.
                    if isinstance(backref, tuple):
                        backref = backref[0]

                    other_relationship = relationship.mapper.get_property(backref)
                    if other_relationship.cascade.delete_orphan:

                        # Is this instance an orphan?
                        if getattr(instance, relationship.key) is None:
                            return True

        return False

    def record_rattail_change(self, session, instance, type_='dirty'):
        """
        Record a change record in the database.

        If ``instance`` represents a change in which we are interested, then
        this method will create (or update) a :class:`rattail.db.model.Change`
        record.

        :returns: ``True`` if a change was recorded, or ``False`` if it was
           ignored.
        """
        # TODO: this check is now redundant due to `ignore_object()`
        # No need to record changes for changes.
        if isinstance(instance, (model.Change, model.DataSyncChange)):
            return False

        # No need to record changes for batch data.
        if isinstance(instance, (model.BatchMixin, model.BatchRowMixin)):
            return False

        # no need to record changes for email attempts
        if isinstance(instance, model.EmailAttempt):
            return False

        # Ignore instances which don't use UUID.
        if not hasattr(instance, 'uuid'):
            return False

        # Provide an UUID value, if necessary.
        self.ensure_uuid(instance)

        # Record the change.
        self.record_change(session, quiet=True,
                           class_name=instance.__class__.__name__,
                           instance_uuid=instance.uuid,
                           deleted=type_ == 'deleted')
        log.debug("recorded change for {} {} {}: {}".format(
            type_, instance.__class__.__name__, instance.uuid, instance))
        return True

    def record_change(self, session, quiet=False, **kwargs):
        """
        Record a change, by creating a new change record in the session.
        """
        session.add(model.Change(**kwargs))
        if not quiet:
            log.debug("recorded {} for {} with key: {}".format(
                'deletion' if kwargs.get('deleted') else 'change',
                kwargs.get('class_name'),
                kwargs.get('object_key', kwargs.get('instance_uuid'))))

    def ensure_uuid(self, instance):
        """
        Ensure the given instance has a UUID value.

        This uses the following logic:

        * If the instance already has a UUID, nothing will be done.

        * If the instance contains a foreign key to another table, then that
          relationship will be traversed and the foreign object's UUID will be used
          to populate that of the instance.

        * Otherwise, a new UUID will be generated for the instance.
        """

        if instance.uuid:
            return

        mapper = object_mapper(instance)
        if not mapper.columns['uuid'].foreign_keys:
            instance.uuid = get_uuid()
            return

        for prop in mapper.iterate_properties:
            if (isinstance(prop, RelationshipProperty)
                and len(prop.remote_side) == 1
                and list(prop.remote_side)[0].key == 'uuid'):

                foreign_instance = getattr(instance, prop.key)
                if foreign_instance:
                    self.ensure_uuid(foreign_instance)
                    instance.uuid = foreign_instance.uuid
                    return

        instance.uuid = get_uuid()
        log.error("unexpected scenario; generated new UUID for instance: {0}".format(repr(instance)))


class ChangeRecorderExtension(SessionExtension):
    """
    Session extension for recording changes.

    .. note::
       This is only used when the installed SQLAlchemy version is old enough
       not to support the new event interfaces.
    """

    def __init__(self, recorder):
        self.recorder = recorder

    def before_flush(self, session, flush_context, instances):
        self.recorder(session, flush_context, instances)



# -*- coding: utf-8; -*-
"""add Role.notes

Revision ID: 6bbcd107cb35
Revises: be2cb8322066
Create Date: 2020-03-23 19:36:16.940525

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '6bbcd107cb35'
down_revision = 'be2cb8322066'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # role
    op.add_column('role', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('role_version', sa.Column('notes', sa.Text(), autoincrement=False, nullable=True))


def downgrade():

    # role
    op.drop_column('role_version', 'notes')
    op.drop_column('role', 'notes')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete handlers - base class
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa


class Autocompleter(object):
    """
    Base class and partial default implementation for autocomplete
    handlers.  It is expected that all autocomplete handlers will
    ultimately inherit from this base class, therefore it defines the
    implementation "interface" loosely speaking.  Custom autocomplete
    handlers are welcome to supplement or override this as needed, and
    in fact must do so for certain aspects.

    .. attribute: autocompleter_key
       The key indicates what "type" of autocompleter this is.  It
       should be a string, e.g. ``'products'``.  It will generally
       correspond to the route names used in Tailbone, though not
       always.  
    """
    autocompleter_key = None

    def __init__(self, config):
        if not self.autocompleter_key:
            raise NotImplementedError("You must define `autocompleter_key` "
                                      "attribute for handler class: {}".format(
                                          self.__class__))
        self.config = config
        self.app = self.config.get_app()
        self.enum = config.get_enum()
        self.model = config.get_model()

    def get_model_class(self):
        return self.model_class

    @property
    def autocomplete_fieldname(self):
        raise NotImplementedError("You must define `autocomplete_fieldname` "
                                  "attribute for handler class: {}".format(
                                      self.__class__))

    def autocomplete(self, session, term, **kwargs):
        """
        The main reason this class exists.  This method accepts a
        ``term`` (string) argument and will return a sequence of
        matching results.
        """
        term = self.prepare_autocomplete_term(term)
        if not term:
            return []

        results = self.get_autocomplete_data(session, term)
        return [{'label': self.autocomplete_display(x),
                 'value': self.autocomplete_value(x)}
                for x in results]

    def prepare_autocomplete_term(self, term, **kwargs):
        """
        If necessary, massage the incoming search term for use with
        the autocomplete query.
        """
        return term

    def get_autocomplete_data(self, session, term, **kwargs):
        """
        Collect data for all matching results, for the given search term.
        """
        query = self.make_autocomplete_query(session, term)
        return query.all()

    def make_autocomplete_query(self, session, term, **kwargs):
        """
        Build the complete query from which to obtain search results.
        """
        # we are querying one table (and column) primarily
        model_class = self.get_model_class()
        query = session.query(model_class)

        # filter according to business logic etc. if applicable
        query = self.filter_autocomplete_query(session, query)

        # filter according to search term(s)
        column = getattr(model_class, self.autocomplete_fieldname)
        criteria = [column.ilike('%{}%'.format(word))
                    for word in term.split()]
        query = query.filter(sa.and_(*criteria))

        # sort results by something meaningful
        query = query.order_by(column)
        return query

    def filter_autocomplete_query(self, session, query, **kwargs):
        return query

    def autocomplete_display(self, obj):
        return getattr(obj, self.autocomplete_fieldname)

    def autocomplete_value(self, obj):
        return obj.uuid



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data model for "new product" batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import (Base, BatchMixin, ProductBatchRowMixin,
                              Vendor, Department, Subdepartment,
                              Category, Family, ReportCode, Brand)
from rattail.db.core import filename_column


class NewProductBatch(BatchMixin, Base):
    """
    Primary data model for new product batches.
    """
    __tablename__ = 'batch_newproduct'
    __batchrow_class__ = 'NewProductBatchRow'
    batch_key = 'newproduct'
    model_title = "New Product Batch"
    model_title_plural = "New Product Batches"

    input_filename = filename_column(nullable=True, doc="""
    Base name of the input data file, if applicable.
    """)


class NewProductBatchRow(ProductBatchRowMixin, Base):
    """
    Row of data within a new product batch.
    """
    __tablename__ = 'batch_newproduct_row'
    __batch_class__ = NewProductBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='batch_newproduct_row_fk_vendor'),
            sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'],
                                    name='batch_newproduct_row_fk_department'),
            sa.ForeignKeyConstraint(['subdepartment_uuid'], ['subdepartment.uuid'],
                                    name='batch_newproduct_row_fk_subdepartment'),
            sa.ForeignKeyConstraint(['category_uuid'], ['category.uuid'],
                                    name='batch_newproduct_row_fk_category'),
            sa.ForeignKeyConstraint(['family_uuid'], ['family.uuid'],
                                    name='batch_newproduct_row_fk_family'),
            sa.ForeignKeyConstraint(['report_uuid'], ['report_code.uuid'],
                                    name='batch_newproduct_row_fk_report'),
            sa.ForeignKeyConstraint(['brand_uuid'], ['brand.uuid'],
                                    name='batch_newproduct_row_fk_brand'),
        )

    STATUS_OK                           = 1
    STATUS_MISSING_KEY                  = 2
    STATUS_PRODUCT_EXISTS               = 3
    STATUS_VENDOR_NOT_FOUND             = 4
    STATUS_DEPT_NOT_FOUND               = 5
    STATUS_SUBDEPT_NOT_FOUND            = 6
    STATUS_CATEGORY_NOT_FOUND           = 7
    STATUS_FAMILY_NOT_FOUND             = 8
    STATUS_REPORTCODE_NOT_FOUND         = 9

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_MISSING_KEY              : "missing product key",
        STATUS_PRODUCT_EXISTS           : "product already exists",
        STATUS_VENDOR_NOT_FOUND         : "vendor not found",
        STATUS_DEPT_NOT_FOUND           : "department not found",
        STATUS_SUBDEPT_NOT_FOUND        : "subdepartment not found",
        STATUS_CATEGORY_NOT_FOUND       : "category not found",
        STATUS_FAMILY_NOT_FOUND         : "family not found",
        STATUS_REPORTCODE_NOT_FOUND     : "report code not found",
    }

    vendor_id = sa.Column(sa.String(length=15), nullable=True, doc="""
    ID string for the vendor from whom product may be purchased.
    """)

    vendor_uuid = sa.Column(sa.String(length=32), nullable=True)
    vendor = orm.relationship(Vendor, doc="""
    Reference to the vendor which corresponds to :attr:`vendor_id`.  Note that
    this may be ``None`` if the ``vendor_id`` value is missing or unknown.
    """)

    vendor_item_code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Vendor-specific item code (SKU) for the product.
    """)

    department_uuid = sa.Column(sa.String(length=32), nullable=True)
    department = orm.relationship(Department, doc="""
    Reference to the department which corresponds to :attr:`department_number`.
    Note that this may be ``None`` if the ``department_number`` value is
    missing or unknown.
    """)

    subdepartment_uuid = sa.Column(sa.String(length=32), nullable=True)
    subdepartment = orm.relationship(Subdepartment, doc="""
    Reference to the subdepartment which corresponds to
    :attr:`subdepartment_number`.  Note that this may be ``None`` if the
    ``subdepartment_number`` value is missing or unknown.
    """)

    category_code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Category code (string value) for the item.
    """)

    category_uuid = sa.Column(sa.String(length=32), nullable=True)
    category = orm.relationship(Category, doc="""
    Reference to the category record which corresponds to
    :attr:`category_code`.  Note that this may be ``None`` if the
    ``category_code`` value is missing or unknown.
    """)

    family_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Family code (integer value) for the item.
    """)

    family_uuid = sa.Column(sa.String(length=32), nullable=True)
    family = orm.relationship(Family, doc="""
    Reference to the "family code" record which corresponds to
    :attr:`family_code`.  Note that this may be ``None`` if the ``family_code``
    value is missing or unknown.
    """)

    report_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Report code (integer value) for the item.
    """)

    report_uuid = sa.Column(sa.String(length=32), nullable=True)
    report = orm.relationship(ReportCode, doc="""
    Reference to the "report code" record which corresponds to
    :attr:`report_code`.  Note that this may be ``None`` if the ``report_code``
    value is missing or unknown.
    """)

    brand_uuid = sa.Column(sa.String(length=32), nullable=True)
    brand = orm.relationship(Brand, doc="""
    Reference to the brand which corresponds to :attr:`brand_name`.  Note that
    this may be ``None`` if the ``brand_name`` value is missing or unknown.
    """)

    case_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of units which constitute a "case" of product.  May be a fractional
    quantity, e.g. 17.5 (LB).
    """)

    case_cost = sa.Column(sa.Numeric(precision=10, scale=5), nullable=True, doc="""
    Cost per case of the product.
    """)

    unit_cost = sa.Column(sa.Numeric(precision=10, scale=5), nullable=True, doc="""
    Cost per unit of the product.
    """)

    regular_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Regular price for the item.
    """)

    regular_price_multiple = sa.Column(sa.Integer(), nullable=True, doc="""
    Regular pricing "multiple" for the item; usually is just 1.
    """)

    pack_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Pack price for the item.
    """)

    pack_price_multiple = sa.Column(sa.Integer(), nullable=True, doc="""
    Pack pricing "multiple" for the item.
    """)

    suggested_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    The "suggested" retail price for the item.
    """)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
TrainWreck -> Trainwreck data importing
"""

from __future__ import unicode_literals, absolute_import

import datetime

from rattail.importing.handlers import FromSQLAlchemyHandler, ToSQLAlchemyHandler
from rattail.importing.sqlalchemy import FromSQLAlchemySameToSame
from rattail.trainwreck.db import Session as TrainwreckSession
from rattail.time import localtime, make_utc
from rattail.util import OrderedDict
from rattail.trainwreck.importing import model
from rattail.trainwreck.importing.util import ToOrFromTrainwreck


class FromTrainwreckHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which have a Trainwreck DB as data source on the host side.
    """
    host_title = "Trainwreck"

    def make_host_session(self):
        return TrainwreckSession()


class ToTrainwreckHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target a Trainwreck DB on the local side.
    """
    local_title = "Trainwreck"

    def make_session(self):
        return TrainwreckSession()


class TrainwreckImportExportBase(FromTrainwreckHandler, ToTrainwreckHandler):
    """
    Shared base class for Trainwreck <-> Trainwreck handlers
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Transaction'] = TransactionImporter
        importers['TransactionItem'] = TransactionItemImporter
        return importers


class FromTrainwreckToTrainwreckImport(TrainwreckImportExportBase):
    """
    Handler for Trainwreck (other) -> Trainwreck (local) data import.

    .. attribute:: direction

       Value is ``'import'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    dbkey = 'host'
    local_title = "Trainwreck (default)"

    @property
    def host_title(self):
        return "Trainwreck ({})".format(self.dbkey)

    def make_host_session(self):
        return TrainwreckSession(bind=self.config.trainwreck_engines[self.dbkey])


class FromTrainwreckToTrainwreckExport(TrainwreckImportExportBase):
    """
    Handler for Trainwreck (local) -> Trainwreck (other) data export.

    .. attribute:: direction

       Value is ``'export'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    direction = 'export'
    host_title = "Trainwreck (default)"

    @property
    def local_title(self):
        return "Trainwreck ({})".format(self.dbkey)

    def make_session(self):
        return TrainwreckSession(bind=self.config.trainwreck_engines[self.dbkey])


class FromTrainwreck(FromSQLAlchemySameToSame):
    """
    Base class for Trainwreck -> Trainwreck data importers.
    """


class TransactionImporter(FromTrainwreck, model.TransactionImporter):
    """
    Base class for Transaction data importer
    """

    def query(self):
        query = super(TransactionImporter, self).query()
        query = self.filter_date_range(query)
        return query

    def filter_date_range(self, query):
        return query.filter(self.model_class.end_time >= make_utc(self.start_time))\
                    .filter(self.model_class.end_time < make_utc(self.end_time))


class TransactionItemImporter(FromTrainwreck, model.TransactionItemImporter):
    """
    Base class for Transaction item data importer
    """

    def query(self):
        query = super(TransactionItemImporter, self).query()
        query = self.filter_date_range(query)
        return query

    def filter_date_range(self, query):
        return query.join(self.transaction_class)\
                    .filter(self.transaction_class.end_time >= make_utc(self.start_time))\
                    .filter(self.transaction_class.end_time < make_utc(self.end_time))



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

from rattail.db import model
from rattail.tests import DataTestCase


class TestCustomer(DataTestCase):

    # TODO: this is duplicated in TestPerson
    def test_add_email_address(self):
        customer = model.Customer()
        self.assertEqual(len(customer.emails), 0)
        customer.add_email_address('fred@mailinator.com')
        self.assertEqual(len(customer.emails), 1)
        email = customer.emails[0]
        self.assertEqual(email.type, 'Home')

        customer = model.Customer()
        self.assertEqual(len(customer.emails), 0)
        customer.add_email_address('fred@mailinator.com', type='Work')
        self.assertEqual(len(customer.emails), 1)
        email = customer.emails[0]
        self.assertEqual(email.type, 'Work')

    # TODO: this is duplicated in TestPerson
    def test_add_phone_number(self):
        customer = model.Customer()
        self.assertEqual(len(customer.phones), 0)
        customer.add_phone_number('417-555-1234')
        self.assertEqual(len(customer.phones), 1)
        phone = customer.phones[0]
        self.assertEqual(phone.type, 'Home')

        customer = model.Customer()
        self.assertEqual(len(customer.phones), 0)
        customer.add_phone_number('417-555-1234', type='Work')
        self.assertEqual(len(customer.phones), 1)
        phone = customer.phones[0]
        self.assertEqual(phone.type, 'Work')



# -*- coding: utf-8; -*-
"""add customer number sequence

Revision ID: e838d0b930fc
Revises: 3ed719791439
Create Date: 2019-10-22 16:10:32.699424

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'e838d0b930fc'
down_revision = u'3ed719791439'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


# we are creating this in the hopes that it may be useful, but we are *not*
# wiring it up, e.g. to provide default values for the Customer.number field -
# you must do that within your own app logic if you want that behavior
customer_number_seq = sa.Sequence('customer_number_seq')


def upgrade():
    from sqlalchemy.schema import CreateSequence

    # customer_number_seq
    op.execute(CreateSequence(customer_number_seq))


def downgrade():
    from sqlalchemy.schema import DropSequence

    # customer_number_seq
    op.execute(DropSequence(customer_number_seq))



# -*- coding: utf-8; -*-
"""add batch_delproduct_row.present_in_scale

Revision ID: 387ba7573623
Revises: 82d3c2d9830d
Create Date: 2021-02-12 17:23:29.947961

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '387ba7573623'
down_revision = '82d3c2d9830d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # department
    op.add_column('department', sa.Column('allow_product_deletions', sa.Boolean(), nullable=True))
    op.add_column('department_version', sa.Column('allow_product_deletions', sa.Boolean(), autoincrement=False, nullable=True))

    # batch_delproduct_row
    op.add_column('batch_delproduct_row', sa.Column('present_in_scale', sa.Boolean(), nullable=True))


def downgrade():

    # batch_delproduct_row
    op.drop_column('batch_delproduct_row', 'present_in_scale')

    # department
    op.drop_column('department_version', 'allow_product_deletions')
    op.drop_column('department', 'allow_product_deletions')
